
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.errors import AnsibleError, AnsibleFilterError
from ansible.module_utils._text import to_text

try:
    import time
    import hashlib
    from datetime import datetime, timedelta
    from OpenSSL import crypto
    HAS_LIB = True
except ImportError:
    HAS_LIB = False

def sign(csr, ca, key, days=30, digest="sha256", serial=None, keypass=None, current_cert=None, renew_days=10, root=None):
    """
    sign a certificate signing request with an intermediate CA certificate and its key. Will also renew an existing
    certificate is the currently installed is provided and the expiry is in renewDays.
    :param csr: The certificate signing request
    :param ca: The (intermediate) certificate authority that will sign this request
    :param key: The private key to be used for signing
    :param days: The days that the certificate will be valid
    :param digest: The algorithm to use for digest, which defaults to sha256
    :param serial: The serial number if any, or a time-dependant serial will be used.
    :param keypass: The password used to encrypt the private key, which is OpenSSL implementation specific.
    :param current_cert: The certificate currently installed, to asses if renewal is required
    :param renew_days: Threashold to decide if renew is required.
    :return: The certificate in PEM format
    """

    if not HAS_LIB:
        raise AnsibleError('You need to install "pyOpenSSL" prior to running '
                           'ca_sign filter')

    try:
        ca_cert = crypto.load_certificate(crypto.FILETYPE_PEM, ca)
    except Exception as e:
        raise AnsibleFilterError("Could not load ca: %s " % e)

    try:
        cert_req = crypto.load_certificate_request(crypto.FILETYPE_PEM, csr)
    except Exception as e:
        raise AnsibleFilterError("Could not load csr: %s " % e)
    try:
        ca_key = crypto.load_privatekey(crypto.FILETYPE_PEM, key, keypass.encode('UTF-8','strict'))
    except Exception as e:
        raise AnsibleFilterError("Could not load key: %s" % e)

    if current_cert is not None:
        # we will check if the current certificate is still up to date.
        try:
            current_cert = crypto.load_certificate(crypto.FILETYPE_PEM, current_cert)
        except Exception as e:
            raise AnsibleFilterError("Could not load current certificate: %s " % e)

        if is_valid_certificate(current_cert, cert_req, renew_days):
            # ensure that the certificate was signed by this ca (ca could have expired and cert still be valid)
            if root:
                try:
                    root = crypto.load_certificate(crypto.FILETYPE_PEM, root)
                except Exception as e:
                    raise AnsibleFilterError("Could not load root certificate: %s " % e)

                if verify(current_cert, ca_cert, root):
                    # the cert is up-to-date and validates, return it
                    return to_text(crypto.dump_certificate(crypto.FILETYPE_PEM, current_cert))
            else:
                # the current certificate is up to date, just return it.
                return to_text(crypto.dump_certificate(crypto.FILETYPE_PEM, current_cert))

    # Certificate generation or renewal:
    digest = "sha256"
    not_after = days * 24 * 60 * 60
    cert = crypto.X509()
    cert.set_version(2)
    if serial is None:
        # The serial number is the md5 of now and the certificate subject as integer
        shash = hashlib.md5()
        shash.update(str(time.time()).encode())
        shash.update(str(cert_req.get_subject()).encode())
        serial = int(shash.hexdigest(),16)
    cert.set_serial_number(serial)
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(not_after)
    cert.set_issuer(ca_cert.get_subject())
    cert.set_subject(cert_req.get_subject())
    cert.set_pubkey(cert_req.get_pubkey())
    cert.add_extensions(cert_req.get_extensions())
    cert.sign(ca_key, digest)
    return to_text(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))


def is_valid_certificate(current_cert, cert_req, renew_days):

    # decide if the renewal is pertinent
    not_after = current_cert.get_notAfter().decode()
    expiry_date = datetime.strptime(not_after, "%Y%m%d%H%M%SZ")

    auto_renew_date = expiry_date - timedelta(days=renew_days)

    if datetime.now() > auto_renew_date:
        # expired, by date
        return False

    if cert_req.get_subject() != current_cert.get_subject():
        # the certificate subject is not current
        return False

    if len(cert_req.get_extensions()) != current_cert.get_extension_count():
        # extension count is not the same
        return False

    for extn in range(0, current_cert.get_extension_count()):
        if cert_req.get_extensions()[extn].get_data() != current_cert.get_extension(extn).get_data():
            # the context of the extension has changed
            return False

    return True


def verify(cert, ca, root=None):
    store = crypto.X509Store()
    store.add_cert(ca)
    if root:
        store.add_cert(root)
    store_ctx = crypto.X509StoreContext(store, cert)
    try:
        store_ctx.verify_certificate()
        return True
    except crypto.X509StoreContextError:
        return False


class FilterModule(object):
    # Ansible hbconf filters

    def filters(self):
        return {
            # jinja2 overrides
            'ca_sign': sign,
        }
